"""ClockWidget provides a clock widget for the main application window"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from datetime import datetime
from typing import Any

from PySide6.QtCore import Slot
from PySide6.QtWidgets import QLCDNumber
from vistutils.fields import IntField, Wait, unParseArgs

from vistside.core import TimerField, Precise
from vistside.widgets import BaseWidget, LCDField, BaseLayoutField


class ClockWidget(BaseWidget):
  """ClockWidget provides a clock widget for the main application window"""

  quartz = TimerField(100, Precise, singleShot=True)
  baseLayout = BaseLayoutField('horizontal')
  sevenSegment = LCDField(8, )
  prevTime = IntField(0, )

  @staticmethod
  def timeDict() -> dict[str, int]:
    """Returns a dictionary with the current time"""
    return {
      "hours"  : datetime.now().hour,
      "minutes": datetime.now().minute,
      "seconds": datetime.now().second,
    }

  @Slot()
  def showTime(self, ) -> None:
    """Shows the current time"""
    timeDict = self.timeDict()
    h, m, s = timeDict['hours'], timeDict['minutes'], timeDict['seconds']
    if self.prevTime - s:
      self.sevenSegment.display('%02d:%02d:%02d' % (h, m, s))
      self.prevTime = s

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the ClockWidget"""
    BaseWidget.__init__(self, *args, **kwargs)
    self.quartz.timeout.connect(self.showTime)
    self.quartz.start()
    self.showTime()

  def initUI(self, ) -> None:
    """Initializes the UI"""
    self.baseLayout.setContentsMargins(0, 0, 0, 0)
    self.sevenSegment.setSegmentStyle(QLCDNumber.SegmentStyle.Filled)
    self.setStyleSheet("""
    QLCDNumber {
      color: black;
      background: rgba(255, 255, 255, 0.2); 
      border: 1px solid black;
      border-radius: 4px;
      padding: 0px;
      margin: 0px;
    }
    """)
    self.baseLayout.addWidget(self.sevenSegment)
    self.setLayout(self.baseLayout)

  def connectActions(self, ) -> None:
    """Connects the actions"""
    self.quartz.timeout.connect(self.showTime)

  @classmethod
  def getDefault(cls, *args, **kwargs) -> ClockWidget:
    """Returns the default ClockWidget"""
    clockWidget = cls()
    clockWidget.apply((args, kwargs))
    return clockWidget

  def apply(self, value: Any) -> ClockWidget:
    """Applies the value to the field"""
    args, kwargs = unParseArgs(value)
    return self


class ClockField(Wait):
  """ClockField provides a widget showing the time"""

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the descriptor"""
    Wait.__init__(self, ClockWidget, *args, **kwargs)



"""LCDField provides a descriptor for QLCDNumber objects."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any, Self

from PySide6.QtWidgets import QLCDNumber
from vistutils.text import stringList
from vistutils.waitaminute import typeMsg

from vistutils.fields import Wait, unParseArgs


class LCDNumber(QLCDNumber):
  """LCDNumber provides a QLCDNumber object."""

  def __init__(self, *args, **kwargs) -> None:
    QLCDNumber.__init__(self, *args, **kwargs)

  @classmethod
  def getDefault(cls, *args, **kwargs) -> Any:
    """Returns the default value for the field."""
    defVal = cls()
    defVal.apply((args, kwargs))
    return defVal

  def apply(self, value: Any) -> Self:
    """Applies the value to the field."""
    args, kwargs = unParseArgs(value)
    numDigs = None
    digitsKey = stringList("""digits, numDigits, num, number, n""")
    for key in digitsKey:
      if key in kwargs and numDigs is None:
        val = kwargs.get(key)
        if isinstance(key, int):
          numDigs = val
          break
        e = typeMsg(key, val, int)
        raise TypeError(e)
    else:
      for arg in args:
        if isinstance(arg, int) and numDigs is None:
          numDigs = arg
          break
    if isinstance(numDigs, int):
      self.setDigitCount(numDigs)
      return self


class LCDField(Wait):
  """LCDField provides a descriptor for QLCDNumber objects."""

  def __init__(self, *args, **kwargs) -> None:
    Wait.__init__(self, LCDNumber, *args, **kwargs)



"""EditMenu provides the edit menu of the main application window"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from vistside.windows.actions import AbstractMenu

if TYPE_CHECKING:
  pass


class EditMenu(AbstractMenu):
  """EditMenu provides the edit menu of the main application window"""

  def setupActions(self, ) -> None:
    """setupActions creates the actions for the menu."""
    self.parent().selectAllAction = self.addAction('Select All', )
    self.parent().cutAction = self.addAction('Cut', )
    self.parent().copyAction = self.addAction('Copy', )
    self.parent().pasteAction = self.addAction('Paste', )
    self.parent().undoAction = self.addAction('Undo', )
    self.parent().redoAction = self.addAction('Redo', )



"""Getter-function for the icon of the application."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os

from PySide6.QtGui import QIcon, QPixmap


def _getPath(actionName: str) -> str:
  """Parses the file path"""
  here = os.path.dirname(os.path.abspath(__file__))
  name = actionName.lower().replace(' ', '_')
  name = os.path.splitext(name)[0]
  fileName = '%s.png' % name
  filePath = os.path.join(here, fileName)
  if os.path.exists(filePath):
    return filePath
  return _getPath('risitas')


def getPixmap(actionName: str) -> QIcon:
  """Returns the icon of the given action."""
  filePath = _getPath(actionName)
  return QPixmap(filePath)


def getIcon(actionName: str) -> QIcon:
  """Returns the icon of the given action."""
  return QIcon(getPixmap(actionName))



"""The parseParent function parses positional arguments and returns the
first instance of QWidget encountered if any."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Optional

from PySide6.QtWidgets import QWidget


def parseParent(*args) -> Optional[QWidget]:
  """The parseParent function parses positional arguments and returns the
  first instance of QWidget encountered if any."""

  for arg in args:
    if isinstance(arg, QWidget):
      return arg



"""WhiteNoisinator class creates white noise"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import numpy as np
from icecream import ic

ic.configureOutput(includeContext=True)


def generateWhiteNoise(sample_length, fs) -> np.ndarray:
  """
  Generates white noise using Inverse Fourier Transform.

  Args:
      sample_length: The desired length of the white noise signal.
      fs: The sampling frequency.

  Returns:
      A NumPy array containing the white noise signal.
  """

  # Create random phases with uniform distribution
  phases = np.random.rand(sample_length) * 2 * np.pi

  # Constant amplitude spectrum
  amplitude = np.ones(sample_length)

  # Combine phases and amplitude
  spectrum = amplitude * np.exp(1j * phases)

  # Perform inverse Fourier Transform (IFFT)

  # Normalize the signal (optional)
  # white_noise = white_noise / np.std(white_noise)

  return np.real(np.fft.ifft(spectrum))


def linearInterpolation(x1, x2, y1, y2, x) -> np.ndarray | float:
  """Performs linear interpolation between two points."""
  return y1 + (x - x1) * (y2 - y1) / (x2 - x1)


def interpolateWhiteNoise(desired_length,
                          fs,
                          initial_sample_length=1000) -> np.ndarray:
  """
  Generates white noise and interpolates to a longer signal length.

  Args:
      desired_length: The desired length of the final white noise signal.
      fs: The sampling frequency.
      initial_sample_length: The number of samples for the initial white
      noise generation.

  Returns:
      A NumPy array containing the interpolated white noise signal.
  """

  # Generate initial white noise
  initNoise = generateWhiteNoise(initial_sample_length, fs)
  thisNoise = initNoise[~np.isnan(initNoise)]

  # Define interpolation function (choose your preferred method)

  # Upsample to desired length
  thisNoise = np.zeros(desired_length)
  for i in range(desired_length - 1):
    x = i / (desired_length - 1) * (
        initial_sample_length - 1)  # Scaled x value
    x1 = int(np.floor(x))
    x2 = int(np.ceil(x))
    y1 = initNoise[x1]
    y2 = initNoise[x2]
    thisNoise[i] = linearInterpolation(x1, x2, y1, y2, x)

  thisNoise[-1] = initNoise[-1]  # Set the last element directly
  # Normalize the signal
  ic(np.min(thisNoise), np.max(thisNoise))
  return thisNoise[~np.isnan(thisNoise)]

#
# # Example usage
# fs = 1000  # Sampling frequency
# desired_length = 10000  # Desired length of the final signal
#
# long_noise = generate_and_interpolate_white_noise(desired_length, fs)
#
# # Further processing (optional)



"""VerticalLayout wraps the QVBoxLayout."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QVBoxLayout

from vistside.core import parseParent


class VerticalLayout(QVBoxLayout):
  """VerticalLayout wraps the QVBoxLayout."""

  def __init__(self, *args, **kwargs) -> None:
    parent = parseParent(*args)
    QVBoxLayout.__init__(self, parent)
    self.setContentsMargins(2, 2, 2, 2, )
    self.setSpacing(2)



"""BaseLayout wraps the basic layout."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtWidgets import QGridLayout, QHBoxLayout, QVBoxLayout, QLayout
from vistutils.fields import Wait
from vistutils.text import stringList


class BaseGrid(QGridLayout):
  """The BaseLayout class is a layout of widgets."""

  def __init__(self, *args, **kwargs) -> None:
    """Create a new BaseLayout."""
    QGridLayout.__init__(self, *args, **kwargs)

  @classmethod
  def getDefault(cls, *args, **kwargs) -> BaseGrid:
    """Returns the default value for the field."""
    return cls()


class BaseHBox(QHBoxLayout):
  """The BaseLayout class is a layout of widgets."""

  def __init__(self, *args, **kwargs) -> None:
    """Create a new BaseLayout."""
    QHBoxLayout.__init__(self, *args, **kwargs)

  @classmethod
  def getDefault(cls, *args, **kwargs) -> BaseHBox:
    """Returns the default value for the field."""
    return cls()


class BaseVBox(QVBoxLayout):
  """The BaseLayout class is a layout of widgets."""

  def __init__(self, *args, **kwargs) -> None:
    """Create a new BaseLayout."""
    QVBoxLayout.__init__(self, *args, **kwargs)

  @classmethod
  def getDefault(cls, *args, **kwargs) -> BaseVBox:
    """Returns the default value for the field."""
    return cls()


class BaseLayoutField(Wait):
  """BaseLayoutField is a field for a BaseLayout."""

  def __init__(self, *args, **kwargs) -> None:
    layoutType = BaseGrid
    layoutKeys = stringList("""layout, layoutType, layoutMode""")
    gridKeys = stringList("""grid, gridLayout, gridMode""")
    hBoxKeys = stringList("""horizontal, hBox, hBoxLayout, hBoxMode""")
    vBoxKeys = stringList("""vertical, vBox, vBoxLayout, vBoxMode""")
    for key in layoutKeys:
      if key in kwargs:
        val = kwargs.get(key)
        if isinstance(val, str):
          if val.lower() in gridKeys:
            layoutType = BaseGrid
          elif val.lower() in hBoxKeys:
            layoutType = BaseHBox
          elif val.lower() in vBoxKeys:
            layoutType = BaseVBox
          else:
            continue
          break
        elif isinstance(val, type):
          if issubclass(val, QLayout):
            layoutType = val
            break
    else:
      for arg in args:
        if isinstance(arg, type):
          if issubclass(arg, QLayout):
            layoutType = arg
            break
      else:
        layoutType = BaseGrid
    Wait.__init__(self, layoutType, *args, **kwargs)



"""AlignmentEnum specifies alignments"""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any, Self

from enum import Enum

from PySide6.QtCore import QRect, QRectF, Qt, QPoint


class AlignmentEnum(Enum):
  """Specifies alignments"""

  @classmethod
  def prime_factors(cls, n: int) -> list:
    """Return the prime factors of a positive integer"""
    if n == 1:
      return []
    if n < 0:
      return [-1, *cls.prime_factors(-n)]
    if n == 0:
      raise ValueError('0 has no prime factors')
    i = 2
    factors = []
    while i * i <= n:
      if n % i:
        i += 1
      else:
        n //= i
        factors.append(i)
    if n > 1:
      factors.append(n)
    return factors

  LEFT = 2
  RIGHT = 3
  TOP = 5
  BOTTOM = 7
  TOP_LEFT = 2 * 5
  TOP_RIGHT = 3 * 5
  BOTTOM_LEFT = 2 * 7
  BOTTOM_RIGHT = 3 * 7

  def __str__(self) -> str:
    """String Representation"""
    return self.name.lower().replace('_', ', ')

  def __add__(self, other: Any) -> Self:
    """Addition"""
    if isinstance(other, AlignmentEnum):
      return AlignmentEnum(self.value * other.value)
    if isinstance(other, int):
      return AlignmentEnum(self.value * other)
    try:
      return self + AlignmentEnum(other)
    except ValueError as valueError:
      raise TypeError from valueError

  def apply(self, sourceRect: QRect, targetRect: QRect) -> QRect:
    """Moves sourceRect to align with targetRect as defined by the enum"""
    sourceSize = sourceRect.size()
    targetSize = targetRect.size()
    sourceHeight, sourceWidth = sourceSize.height(), sourceSize.width()
    targetHeight, targetWidth = targetSize.height(), targetSize.width()
    targetCenter = targetRect.center()
    targetX, targetY = targetCenter.x(), targetCenter.y()
    sourceCenter = sourceRect.center()
    sourceX, sourceY = sourceCenter.x(), sourceCenter.y()
    if self.value % 2 == 0:  # Left
      left = 0
      right = sourceWidth
    elif self.value % 3 == 0:
      left = targetWidth - sourceWidth
      right = targetWidth
    else:
      left = targetWidth / 2 - sourceWidth / 2
      right = targetWidth / 2 + sourceWidth / 2
    if self.value % 5 == 0:  # Top
      top = 0
      bottom = sourceHeight
    elif self.value % 7 == 0:
      top = targetHeight - sourceHeight
      bottom = targetHeight
    else:
      top = targetHeight / 2 - sourceHeight / 2
      bottom = targetHeight / 2 + sourceHeight / 2
    left += targetRect.left()
    right += targetRect.left()
    top += targetRect.top()
    bottom += targetRect.top()
    return QRect(QPoint(left, top), QPoint(right, bottom))

  def toQt(self) -> Qt.AlignmentFlag:
    """Returns the equivalent Qt.AlignmentFlag"""
    if self == AlignmentEnum.LEFT:
      return Qt.AlignmentFlag.AlignLeft
    if self == AlignmentEnum.RIGHT:
      return Qt.AlignmentFlag.AlignRight
    if self == AlignmentEnum.TOP:
      return Qt.AlignmentFlag.AlignTop
    if self == AlignmentEnum.BOTTOM:
      return Qt.AlignmentFlag.AlignBottom
    if self == AlignmentEnum.TOP_LEFT:
      return Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft
    if self == AlignmentEnum.TOP_RIGHT:
      return Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignRight
    if self == AlignmentEnum.BOTTOM_LEFT:
      return Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignLeft
    if self == AlignmentEnum.BOTTOM_RIGHT:
      return Qt.AlignmentFlag.AlignBottom | Qt.AlignmentFlag.AlignRight
    raise ValueError



"""The 'actions' module provides menus and classes for the main
application window. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._abstract_menu import AbstractMenu
from ._files_menu import FilesMenu
from ._edit_menu import EditMenu
from ._help_menu import HelpMenu
from ._debug_menu import DebugMenu
from ._main_status_bar import MainStatusBar



"""The widgets module provides widget classes for the GUI framework."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._base_widget import BaseWidget
from ._base_layout_field import BaseLayoutField
from ._fill_widget import FillWidget
from ._label_widget import LabelWidget, LabelField
from ._lcd_field import LCDField
from ._clock_widget import ClockWidget, ClockField
from ._abstract_dialog_field import AbstractDialogField
from ._confirm_dialog import ConfirmDialog
from ._plot_widget import PlotWidget
from ._dynamic_plot_widget import DynamicPlotWidget



"""ToggleButton implements a toggle button that is easy to deactivate but
slow to activate. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QPainter, QPen, QBrush
from PySide6.QtCore import QTimer, Qt, QPointF

from vistside.core import parseParent
from vistside.widgets import (BaseWidget)


class SafetyToggleButton(BaseWidget):
  """SafetyToggleButton implements a toggle button that is easy to
  deactivate, but difficult to activate."""

  __activate_time__ = 1000  # Time in ms to activate
  __time_steps__ = 50  # Number of steps to activate

  def __init__(self, *args, **kwargs) -> None:
    parent = parseParent(*args)
    BaseWidget.__init__(self, parent)
    self.isHigh = False  # Toggle state
    self.progress = 0.0  # Progress of the timer
    self.timer = QTimer(self)  # Timer for delay
    self.timer.setInterval(self.getStepTime())  # Update every 100 ms
    self.timer.timeout.connect(self.updateProgress)
    self.setMinimumSize(100, 100)

  def getStepTime(self) -> int:
    """Getter-function for the step time"""
    return int(round(self.__activate_time__ / self.__time_steps__))

  def getFullTime(self) -> int:
    """Getter-function for the full time"""
    return self.__activate_time__

  def mousePressEvent(self, event) -> None:
    """Starts the timer on mouse press."""
    if event.button() == Qt.LeftButton:
      self.progress = 0.0
      self.timer.start()

  def mouseReleaseEvent(self, event) -> None:
    """Stops the timer and toggles state if held long enough."""
    if self.isHigh:
      self.isHigh = False
      self.timer.stop()
      return self.update()
    if self.progress >= 1.0:
      self.isHigh = not self.isHigh
    self.progress = 0.0
    self.update()

  def updateProgress(self) -> None:
    """Updates the progress and repaints the widget."""
    self.progress += self.getStepTime()
    if self.progress >= self.getFullTime():
      self.timer.stop()
    self.update()

  def getProgression(self, ) -> float:
    """Getter for progression."""
    return min(1.00, self.progress / self.getFullTime())

  def paintEvent(self, event) -> None:
    """Draws the toggle button with progress indication."""
    painter = QPainter(self)
    painter.setRenderHint(QPainter.Antialiasing)

    # Draw the circle around the button
    pen = QPen(Qt.black, 2)
    brush = QBrush(Qt.white)
    painter.setPen(pen)
    radius = min(self.width(), self.height()) / 2 - 10
    center = QPointF(self.width() / 2, self.height() / 2)
    painter.drawEllipse(center, radius, radius)

    # Draw progress
    if self.getProgression() > 0:
      pen.setColor(Qt.green)
      painter.setPen(pen)
      brush = QBrush(Qt.green)
      painter.setBrush(brush)
      angle = 360 * self.getProgression()
      painter.drawPie(center.x() - radius, center.y() - radius, radius * 2,
                      radius * 2, 90 * 16, -angle * 16)

    # Draw the toggle state
    pen.setColor(Qt.blue if self.isHigh else Qt.red)
    painter.setPen(pen)
    brush.setColor(Qt.blue if self.isHigh else Qt.red)
    painter.setBrush(brush)
    painter.drawEllipse(center, radius / 2, radius / 2)



"""The fontFactory function creates instances of QFont."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtGui import QFont
from PySide6.QtGui import QFontDatabase
from vistutils.text import stringList
from vistutils.parse import maybe


def _getFontFamilies() -> list[str]:
  """Getter-function for font families."""
  return [f.lower() for f in QFontDatabase().families()]


def resolveFontFamily(family: str, **kwargs) -> str:
  """Resolves a font family."""
  families = _getFontFamilies()
  if family.lower() in families:
    return family
  if kwargs.get('strict', True):
    e = """The name '%s' was not recognized as a font family."""
    raise NameError(e)
  return ''


def _createFont(*args, **kwargs) -> QFont:
  """Creates a QFont instance."""
  familyDefault = "Courier"
  sizeDefault = 16
  weightDefault = QFont.Weight.Normal
  familyArg, sizeArg, weightArg = None, None, None
  familyKeys = stringList("""family, fontFamily""")
  sizeKeys = stringList("""size, fontSize""")
  weightKeys = stringList("""weight, fontWeight""")
  for arg in args:
    if isinstance(arg, str):
      family = resolveFontFamily(arg, strict=False)
      if family and familyArg is None:
        familyArg = family
    elif isinstance(arg, int) and sizeArg is None:
      sizeArg = arg
  familyKwarg, sizeKwarg, weightKwarg = None, None, None
  for key in familyKeys:
    if key in kwargs and familyKwarg is None:
      familyKwarg = kwargs[key]
  for key in sizeKeys:
    if key in kwargs and sizeKwarg is None:
      sizeKwarg = kwargs[key]
  for key in weightKeys:
    if key in kwargs and weightKwarg is None:
      weightKwarg = kwargs[key]
  family = maybe(familyKwarg, familyArg, familyDefault)
  size = maybe(sizeKwarg, sizeArg, sizeDefault)
  weight = maybe(weightKwarg, weightArg, weightDefault)
  font = QFont()
  font.setFamily(family)
  font.setPointSize(size)
  font.setWeight(weight)
  return font


def parseFont(*args, **kwargs) -> QFont:
  """Parses a QFont instance."""
  return _createFont(*args, **kwargs)



"""DynamicPlotWidget provides a widget for plotting dynamic data."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import time

import numpy as np
from PySide6.QtGui import QPainter
from vistutils.fields import IntField, Wait

from vistside.core import parseParent, ArrayField, DataRollField
from vistside.widgets import BaseWidget
import sys
from PySide6.QtWidgets import QApplication, QWidget, QVBoxLayout
from PySide6.QtCharts import QChart, QChartView, QLineSeries, QValueAxis
from PySide6.QtCore import Qt, QTimer, QObject, QThread, Signal, Slot


class DynamicPlotWidget(BaseWidget):
  """DynamicPlotWidget provides a widget for plotting dynamic data."""

  dataRoll = DataRollField(128)

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the widget."""
    BaseWidget.__init__(self, *args, **kwargs)
    self.current_index = 0

    # Chart setup
    self.chart = QChart()
    self.series = QLineSeries()
    self.chart.addSeries(self.series)

    axis_x = QValueAxis()  # For the time axis
    axis_y = QValueAxis()  # For the value axis
    self.chart.setAxisX(axis_x, self.series)
    self.chart.setAxisY(axis_y, self.series)

    # Customize axes, titles, etc. as needed (add your code here)

    self.chart_view = QChartView(self.chart)
    self.chart_view.setRenderHint(QPainter.Antialiasing)

    layout = QVBoxLayout()
    layout.addWidget(self.chart_view)
    self.setLayout(layout)

  @Slot(float)
  def callback(self, value) -> None:
    """Appends new data (value, timestamp)."""
    self.dataRoll.append(value)

  @Slot()
  def updateSlot(self) -> None:
    """Updates the plot visualization."""
    self.series.clear()
    self.series.appendNp(self.dataRoll.real, self.dataRoll.imag)
    # self.chart.update()
    self.update()



"""FontField provides a field for selecting a font."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from PySide6.QtGui import QFont
from vistutils.fields import CoreDescriptor

from vistside.core import parseFont, resolveFontFamily


class FontField(CoreDescriptor):
  """The FontField class provides a field for selecting a font."""

  def _instantiate(self, instance: object) -> None:
    """Instantiates the field."""
    font = parseFont(*self._getArgs(), **self._getKwargs())
    setattr(instance, self._getPrivateName(), font)

  def __set__(self, instance: object, value: Any) -> None:
    """Sets the field."""
    pvtName = self._getPrivateName()
    owner = self._getFieldOwner()
    if isinstance(value, str):
      try:
        family = resolveFontFamily(value)
        existingFont = self.__get__(instance, owner)
        existingFont.setFamily(family)
        return setattr(instance, pvtName, existingFont)
      except NameError as nameError:
        e = """Tried setting font to a string, but the string was not 
        recognized as a valid font family name!"""
        raise ValueError(e) from nameError
    if isinstance(value, QFont):
      return setattr(instance, pvtName, value)
    if isinstance(value, int):
      existingFont = self.__get__(instance, owner)
      existingFont.setPointSize(value)
      return setattr(instance, pvtName, existingFont)
    if isinstance(value, QFont.Weight):
      existingFont = self.__get__(instance, owner)
      existingFont.setWeight(value)
      return setattr(instance, pvtName, existingFont)
    e = """The FontField supports setting the font family to a string, 
    the font size to an integer, the font weight to an enum value of 
    QFont.Weight, or the entire font to an instance of QFont. The value 
    received was of type %s.""" % type(value)
    raise TypeError(value)



"""Common names from Qt."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from PySide6.QtCore import Qt, QEvent
from PySide6.QtGui import QMouseEvent
from PySide6.QtWidgets import QSizePolicy

__all__ = []

from vistutils.text import stringList

#  Timer types
TimerType = Qt.TimerType
Precise = Qt.TimerType.PreciseTimer
Coarse = Qt.TimerType.CoarseTimer
VeryCoarse = Qt.TimerType.VeryCoarseTimer
__all__ += stringList("""TimerType, Precise, Coarse, VeryCoarse""")
#  Text flags
TextFlag = Qt.TextFlag
WordWrap = Qt.TextFlag.TextWordWrap
CharWrap = Qt.TextFlag.TextWrapAnywhere
NoWrap = Qt.TextFlag.TextSingleLine
__all__ += stringList("""TextFlag, WordWrap, CharWrap, NoWrap""")
#  Alignments
Left = Qt.AlignmentFlag.AlignLeft
Right = Qt.AlignmentFlag.AlignRight
Top = Qt.AlignmentFlag.AlignTop
Bottom = Qt.AlignmentFlag.AlignBottom
Center = Qt.AlignmentFlag.AlignCenter
HCenter = Qt.AlignmentFlag.AlignHCenter
VCenter = Qt.AlignmentFlag.AlignVCenter
__all__ += stringList(
  """Left, Right, Top, Bottom, Center, HCenter, VCenter""")
#  Pen styles
SolidLine = Qt.PenStyle.SolidLine
DashLine = Qt.PenStyle.DashLine
DotLine = Qt.PenStyle.DotLine
DashDotLine = Qt.PenStyle.DashDotLine
DashDotDotLine = Qt.PenStyle.DashDotDotLine
EmptyLine = Qt.PenStyle.NoPen
RoundCap = Qt.PenCapStyle.RoundCap
FlatCap = Qt.PenCapStyle.FlatCap
SquareCap = Qt.PenCapStyle.SquareCap
__all__ += stringList(
  """SolidLine, DashLine, DotLine, DashDotLine, DashDotDotLine, EmptyLine,
  RoundCap, FlatCap, SquareCap""")
#  Brush styles
SolidFill = Qt.BrushStyle.SolidPattern
__all__ += stringList("""SolidFill""")
#  Size policy
Spread = QSizePolicy.Policy.MinimumExpanding
Fixed = QSizePolicy.Policy.Fixed
Tight = QSizePolicy.Policy.Maximum
__all__ += stringList("""Spread, Fixed, Tight""")
#  Mouse buttons
MouseBtn = Qt.MouseButton
NoBtn = Qt.MouseButton.NoButton
LeftBtn = Qt.MouseButton.LeftButton
RightBtn = Qt.MouseButton.RightButton
MiddleBtn = Qt.MouseButton.MiddleButton
NextBtn = Qt.MouseButton.ForwardButton
BackBtn = Qt.MouseButton.BackButton
__all__ += stringList(
  """NoBtn, MouseBtn, LeftBtn, RightBtn, MiddleBtn, NextBtn, BackBtn""")
#  Mouse events
Mouse = QMouseEvent
BtnDown = QEvent.Type.MouseButtonPress
BtnUp = QEvent.Type.MouseButtonRelease
MouseMove = QEvent.Type.MouseMove
__all__ += stringList("""Mouse, BtnDown, BtnUp, MouseMove""")



"""ConfirmDialog provides a descriptor class that inherits from QObject
enabling it to emit signals. Use the apply decorator on methods that
should require confirmation before being allowed to run. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod
from typing import Callable, Any

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QDialog

from vistside.widgets import AbstractDialogField


class ConfirmDialog(AbstractDialogField):
  """ConfirmDialog provides a descriptor class that inherits from QObject
  enabling it to emit signals. Use the apply decorator on methods that
  should require confirmation before being allowed to run. """

  accepted = Signal()
  rejected = Signal()

  def trustMeBro(self, *args, **kwargs) -> Callable:
    """The most outer function is called in the line above the method. """

    def decorator(callMeMaybe: Callable) -> Callable:
      """The next level receives the decorated function."""

      def wrapper(instance: object, *args2, **kwargs2) -> None:
        """The wrapper replaces the decorated function. The wrapper
        creates a dialog, connects its accepted signal to the decorated
        function. Then opens the dialog. """

        def callception() -> None:
          """Calls the method that should require confirmation."""
          return callMeMaybe(*args2, **kwargs2)

        dialog = self._createDialog(instance, callMeMaybe)
        dialog.accepted.connect(callception)
        dialog.open()

      return wrapper

    return decorator

  def _createDialog(self, instance: Any, callMeMaybe: Callable) -> QDialog:
    """Creates the dialog for the field. Please note that the dialog
    receives only the instance under consideration. """
    insName = str(instance)
    funcName = callMeMaybe.__name__
    msg = """Object: '%s' is asking for permission to invoke callable: 
    '%s'. """ % (insName, funcName)
    title = '%s.%s' % (insName, funcName)
    dialog = QDialog()
    dialog.setWindowTitle(title)
    layout = GridLayout()
    header = LabelWidget(msg)



"""The timerFactory creates the QTimer instances used in the application. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from PySide6.QtCore import QTimer, Qt
from vistutils.parse import maybe
from vistutils.text import stringList

from vistside.core import Precise


def _parseArgs(*args, **kwargs) -> dict:
  """timerFactory creates a QTimer instance. """
  epochKwarg, shotKwarg, timerTypeKwarg = None, None, None
  epochArg, shotArg, timerTypeArg = None, None, None
  epochDefault, shotDefault, timerTypeDefault = 100, False, Precise
  epochKeys = stringList("""epoch, length, interval, time""")
  shotKeys = stringList("""shot, singleShot""")
  notShotKeys = stringList("""repeat, again""")
  timerTypeKeys = stringList("""timerType, precision, accuracy""")
  names = stringList("""epoch, shot, notShot, timerType""")
  keyArgs = dict(epoch=None, timerType=None)
  Keys = [epochKeys, shotKeys, notShotKeys, timerTypeKeys]
  types = [int, bool, bool, Qt.TimerType]
  for (keys, name, type_,) in zip(Keys, names, types):
    for key in keys:
      if key in kwargs:
        if isinstance(kwargs[key], type_):
          keyArgs[name] = kwargs[key]
  if 'shot' in keyArgs and 'notShot' in keyArgs:
    e = 'Both shot and notShot are present in the arguments.'
    raise ValueError(e)
  if 'shot' in keyArgs:
    shotKwarg = keyArgs['shot']
  elif 'notShot' in keyArgs:
    shotKwarg = not keyArgs['notShot']
  epochKwarg = keyArgs['epoch']
  timerTypeKwarg = keyArgs['timerType']
  for arg in args:
    if isinstance(arg, int) and epochArg is None:
      epochArg = arg
    if isinstance(arg, bool) and shotArg is None:
      shotArg = arg
    if isinstance(arg, Qt.TimerType) and timerTypeArg is None:
      timerTypeArg = arg
  epoch = maybe(epochKwarg, epochArg, epochDefault)
  shot = maybe(shotKwarg, shotArg, shotDefault)
  timerType = maybe(timerTypeKwarg, timerTypeArg, timerTypeDefault)
  return dict(epoch=epoch, shot=shot, timerType=timerType)


def parseTimer(*args, **kwargs) -> dict:
  """The timerFactory creates a QTimer instance. """
  parsed = _parseArgs(*args, **kwargs)
  timer = QTimer()
  timer.setInterval(parsed['epoch'])
  timer.setTimerType(parsed['timerType'])
  timer.setSingleShot(parsed['shot'])
  return timer



"""FontPenField provides a static descriptor class for QPen instances for
rendering text. Static means that they are not sensitive to which instance
calls for them, but only whether the field was accessed through and
instance or through the owner. """
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Optional

from PySide6.QtGui import QPen
from vistutils.fields import StaticField
from vistutils.parse import maybe
from vistutils.text import stringList
from vistutils.waitaminute import typeMsg

from vistside.core import SolidLine, parsePen


class FontPenField(StaticField):
  """The FontPenField class provides a descriptor for instances of QPen for
  rendering text. """

  __font_pen__ = None

  @staticmethod
  def _getFallbackPen() -> QPen:
    """Returns a fallback pen."""
    pen = QPen()
    pen.setColor("black")
    pen.setStyle(SolidLine)
    pen.setWidth(1)
    return pen

  @staticmethod
  def _parseKwargs(**kwargs) -> Optional[QPen]:
    """Parses the keyword arguments."""
    penKeys = stringList("""pen, fontPen, font_pen, fontLine, line""")
    for key in penKeys:
      if key in kwargs:
        val = kwargs.get(key)
        if isinstance(val, QPen):
          return val
        e = typeMsg(key, val, QPen)
        raise TypeError(e)

  @staticmethod
  def _parseArgs(*args, ) -> Optional[QPen]:
    """Parses the positional arguments."""
    for arg in args:
      if isinstance(arg, QPen):
        return arg

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the descriptor."""
    penKwarg = self._parseKwargs(**kwargs)
    penArg = self._parseArgs(*args)
    penDefault = self._getFallbackPen()
    fontPen = maybe(penKwarg, penArg, penDefault)
    if isinstance(fontPen, QPen):
      self.__font_pen__ = fontPen
      StaticField.__init__(self, fontPen, )



"""LabelWidget provides a widget for displaying labels. This is intended
for short names or descriptions rather than longer text."""
#  MIT Licence
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Self
from warnings import warn

from PySide6.QtCore import QRect, QPoint, QMargins, QSize, Qt
from PySide6.QtGui import QColor, QPaintEvent, QPainter, QFontMetrics, QFont
from vistutils.text import monoSpace
from vistutils.fields import TextField, Wait, unParseArgs

from vistside.core import FontField, PenField, SolidLine
from vistside.core import resolveFontFamily, NoWrap
from vistside.core import BrushField, SolidFill, Yellow
from vistside.widgets import BaseWidget


class LabelWidget(BaseWidget):
  """LabelWidget provides a widget for displaying labels. This is intended
  for short names or descriptions rather than longer text."""

  innerText = TextField('Label')
  textFont = FontField('Courier', 18)
  fillBrush = BrushField(Yellow, SolidFill)
  borderPen = PenField(QColor(144, 144, 255, 255), 2, SolidLine)
  textFillBrush = BrushField(QColor(255, 255, 255, 63), SolidFill)
  textBorderPen = PenField(QColor(0, 0, 0, 63), 1, SolidLine)
  hAlign = TextField('center')
  vAlign = TextField('center')

  def __init__(self, text: str = None, *args, **kwargs) -> None:
    """Initializes the LabelWidget."""
    BaseWidget.__init__(self, *args, **kwargs)
    self.innerText = 'Label' if text is None else text
    fontMetrics = QFontMetrics(self.textFont)
    text = self.innerText
    textOption = NoWrap
    space = QRect(QPoint(0, 0), QSize(1920, 1080))
    textRect = fontMetrics.boundingRect(space, textOption, text)
    textSize = (textRect + QMargins(8, 1, 8, 1)).size()
    self.setMinimumSize(textSize)

  def paintEvent(self, event: QPaintEvent) -> None:
    """Paints the widget."""
    painter = QPainter()
    painter.begin(self)
    viewRect = painter.viewport()
    b = self.borderPen.width() * 2
    borderMargins = QMargins(b, b, b, b)
    borderRect = viewRect.marginsRemoved(borderMargins)
    borderRect.moveCenter(viewRect.center())
    b = self.textBorderPen.width() * 4
    textBorderMargins = QMargins(b, b, b, b)
    textSpace = borderRect.marginsRemoved(textBorderMargins)
    painter.setRenderHint(QPainter.Antialiasing)
    # # # # # # # # # # # # # # # # #
    #  fill background
    painter.setPen(self.emptyPen)
    painter.setBrush(self.fillBrush)
    painter.drawRect(viewRect)
    # # # # # # # # # # # # # # # # #
    #  draw border
    painter.setPen(self.borderPen)
    painter.setBrush(self.emptyBrush)
    painter.drawRect(borderRect)
    # # # # # # # # # # # # # # # # #
    #  Calculate required size for text
    painter.setFont(self.textFont)
    painter.setPen(self.textPen)
    textRect = painter.boundingRect(textSpace, NoWrap, self.innerText)
    textSize = (textRect + QMargins(8, 1, 8, 1)).size()
    # # # # # # # # # # # # # # # # #
    #  Align text rectangle
    if self.hAlign.lower() == 'center':
      left = textSpace.center().x() - textSize.width() / 2
    elif self.hAlign.lower() == 'right':
      left = textSpace.right() - textSize.width()
    elif self.hAlign.lower() == 'left':
      left = textSpace.left()
    else:
      e = """Invalid horizontal alignment! Expected one of: 'center', 
      'right', 'left', but received: '%s'!""" % self.hAlign
      raise ValueError(monoSpace(e))
    if self.vAlign.lower() == 'center':
      top = textSpace.center().y() - textSize.height() / 2
    elif self.vAlign.lower() == 'bottom':
      top = textSpace.bottom() - textSize.height()
    elif self.vAlign.lower() == 'top':
      top = textSpace.top()
    else:
      e = """Invalid vertical alignment! Expected one of: 'center', 'bottom',
      'top', but received: '%s'!""" % self.vAlign
      raise ValueError(monoSpace(e))
    textTopLeft = QPoint(left, top)
    textRect = QRect(textTopLeft, textSize)
    # # # # # # # # # # # # # # # # #
    #  Fill text background
    painter.setPen(self.emptyPen)
    painter.setBrush(self.textFillBrush)
    painter.drawRect(textRect)
    # # # # # # # # # # # # # # # # #
    #  Draw text border
    painter.setPen(self.textBorderPen)
    painter.setBrush(self.emptyBrush)
    painter.drawRect(textRect)
    # # # # # # # # # # # # # # # # #
    #  Print text
    painter.setPen(self.textPen)
    flags = NoWrap | Qt.AlignmentFlag.AlignHCenter
    painter.drawText(textRect, flags, self.innerText)
    painter.end()

  @classmethod
  def getDefault(cls, *args, **kwargs) -> Self:
    """Creates default"""
    labelWidget = cls()
    return labelWidget.apply((args, kwargs))

  def apply(self, value: tuple) -> Self:
    """Applies values from arguments"""
    args, kwargs = unParseArgs(value)
    warn('Setter not fully implemented!')
    blanks = [None for _ in range(2)]
    strArgs = [*[arg for arg in args if isinstance(arg, str)], *blanks]
    intArgs = [*[arg for arg in args if isinstance(arg, int)], *blanks]
    text, fontFamily = strArgs[:2]
    fontSize = intArgs[0]
    if text is not None:
      self.innerText = text
    if fontFamily is not None:
      fontFamily = resolveFontFamily(fontFamily, strict=False)
      if fontFamily:
        self.textFont = fontFamily
    if fontSize is not None:
      self.textFont = fontSize
    return self


class LabelField(Wait):
  """LabelField is a field for a LabelWidget."""

  def __init__(self, *args, **kwargs) -> None:
    """Initializes the descriptor"""
    Wait.__init__(self, LabelWidget, *args, **kwargs)
